/*
 * Work under Copyright. Licensed under the EUPL.
 * See the project README.md and LICENSE.txt for more information.
 */

package net.dries007.tfc.objects.blocks.stone;

import java.util.EnumMap;
import java.util.HashMap;
import java.util.Map;
import java.util.Random;
import javax.annotation.ParametersAreNonnullByDefault;

import net.minecraft.block.Block;
import net.minecraft.block.BlockFarmland;
import net.minecraft.block.BlockGrassPath;
import net.minecraft.block.SoundType;
import net.minecraft.block.state.IBlockState;
import net.minecraft.init.Items;
import net.minecraft.item.Item;
import net.minecraft.util.BlockRenderLayer;
import net.minecraft.util.EnumFacing;
import net.minecraft.util.math.BlockPos;
import net.minecraft.world.IBlockAccess;
import net.minecraft.world.World;
import net.minecraftforge.common.EnumPlantType;
import net.minecraftforge.common.IPlantable;
import net.minecraftforge.fml.relauncher.Side;
import net.minecraftforge.fml.relauncher.SideOnly;

import mcp.MethodsReturnNonnullByDefault;
import net.dries007.tfc.api.types.Rock;
import net.dries007.tfc.objects.blocks.BlocksTFC;
import net.dries007.tfc.objects.items.rock.ItemRock;
import net.dries007.tfc.util.Helpers;
import net.dries007.tfc.util.OreDictionaryHelper;

@MethodsReturnNonnullByDefault
@ParametersAreNonnullByDefault
public class BlockRockVariant extends Block
{
    private static final Map<Rock, EnumMap<Rock.Type, BlockRockVariant>> TABLE = new HashMap<>();

    public static BlockRockVariant get(Rock rock, Rock.Type type)
    {
        return TABLE.get(rock).get(type);
    }

    public static BlockRockVariant create(Rock rock, Rock.Type type)
    {
        switch (type)
        {
            case RAW:
                return new BlockRockRaw(type, rock);
            case FARMLAND:
                return new BlockFarmlandTFC(type, rock);
            case PATH:
                return new BlockPathTFC(type, rock);
            case GRASS:
            case DRY_GRASS:
            case CLAY_GRASS:
                return new BlockRockVariantConnected(type, rock);
            case SAND:
            case DIRT:
            case CLAY:
            case GRAVEL:
            case COBBLE:
                return new BlockRockVariantFallable(type, rock);
            default:
                return new BlockRockVariant(type, rock);
        }
    }

    public final Rock.Type type;
    public final Rock rock;

    public BlockRockVariant(Rock.Type type, Rock rock)
    {
        super(type.material);

        if (!TABLE.containsKey(rock))
            TABLE.put(rock, new EnumMap<>(Rock.Type.class));
        TABLE.get(rock).put(type, this);

        this.type = type;
        this.rock = rock;
        if (type.isGrass) func_149675_a(true);
        switch (type)
        {
            case BRICKS:
            case RAW:
                func_149672_a(SoundType.field_185851_d);
                func_149711_c(2.0F).func_149752_b(10.0F);
                setHarvestLevel("pickaxe", 0);
                break;
            case COBBLE:
            case SMOOTH:
                func_149672_a(SoundType.field_185851_d);
                func_149711_c(1.5F).func_149752_b(10.0F);
                setHarvestLevel("pickaxe", 0);
                break;
            case SAND:
                func_149672_a(SoundType.field_185855_h);
                func_149711_c(0.5F);
                setHarvestLevel("shovel", 0);
                break;
            case DIRT:
            case PATH:
            case FARMLAND:
                func_149672_a(SoundType.field_185849_b);
                func_149711_c(0.5F);
                setHarvestLevel("shovel", 0);
                break;
            case GRAVEL:
                func_149672_a(SoundType.field_185849_b);
                func_149711_c(0.6F);
                setHarvestLevel("shovel", 0);
                break;
            case CLAY:
            case CLAY_GRASS:
            case GRASS:
            case DRY_GRASS:
                func_149672_a(SoundType.field_185850_c);
                func_149711_c(0.6F);
                setHarvestLevel("shovel", 0);
                break;
        }
        OreDictionaryHelper.registerRockType(this, type, rock);
    }

    public BlockRockVariant getVariant(Rock.Type t)
    {
        return TABLE.get(rock).get(t);
    }

    @Override
    @SideOnly(Side.CLIENT)
    @SuppressWarnings("deprecation")
    public boolean func_176225_a(IBlockState blockState, IBlockAccess world, BlockPos pos, EnumFacing side)
    {
        switch (this.type)
        {
            case PATH:
            case FARMLAND:
                switch (side)
                {
                    case UP:
                        return true;
                    case NORTH:
                    case SOUTH:
                    case WEST:
                    case EAST:
                        IBlockState state = world.func_180495_p(pos.func_177972_a(side));
                        Block block = state.func_177230_c();
                        if (state.func_185914_p()) return false;
                        if (block instanceof BlockFarmland || block instanceof BlockGrassPath) return false;
                        if (block instanceof BlockRockVariant)
                        {
                            switch (((BlockRockVariant) block).type)
                            {
                                case FARMLAND:
                                case PATH:
                                    return false;
                                default:
                                    return true;
                            }
                        }
                        return true;
                    case DOWN:
                        return super.func_176225_a(blockState, world, pos, side);
                }
            default:
                return super.func_176225_a(blockState, world, pos, side);
        }
    }

    @Override
    public void func_180645_a(World world, BlockPos pos, IBlockState state, Random rand)
    {
        if (world.field_72995_K) return;
        if (type.isGrass) Helpers.spreadGrass(world, pos, state, rand);
        super.func_180645_a(world, pos, state, rand);
    }

    @Override
    public Item func_180660_a(IBlockState state, Random rand, int fortune)
    {
        switch (type)
        {
            case RAW:
                return ItemRock.get(rock);
            case CLAY:
            case CLAY_GRASS:
                return Items.field_151119_aD; // todo: own clay or event for clay making?
            default:
                return super.func_180660_a(state, rand, fortune);
            case GRASS:
            case DRY_GRASS:
            case PATH:
                return Item.func_150898_a(get(rock, Rock.Type.DIRT));
        }
    }

    @Override
    public int func_180651_a(IBlockState state)
    {
        return func_176201_c(state);
    }

    @Override
    @SideOnly(Side.CLIENT)
    public BlockRenderLayer func_180664_k()
    {
        return type.isGrass ? BlockRenderLayer.CUTOUT : BlockRenderLayer.SOLID;
    }

    @Override
    public int quantityDropped(IBlockState state, int fortune, Random random)
    {
        switch (type)
        {
            case CLAY:
            case CLAY_GRASS:
                return 4;
            case RAW:
                return 1 + random.nextInt(3);
            default:
                return super.quantityDropped(state, fortune, random);
        }
    }

    @Override
    public boolean canSustainPlant(IBlockState state, IBlockAccess world, BlockPos pos, EnumFacing direction, IPlantable plantable)
    {
        EnumPlantType plantType = plantable.getPlantType(world, pos.func_177972_a(direction));

        switch (plantType)
        {
            case Plains:
                return type == Rock.Type.DIRT || type == Rock.Type.GRASS || type == Rock.Type.DRY_GRASS || type == Rock.Type.CLAY_GRASS;
            case Crop:
                return type == Rock.Type.DIRT || type == Rock.Type.GRASS || type == Rock.Type.FARMLAND || type == Rock.Type.DRY_GRASS;
            case Desert:
                return type == Rock.Type.SAND;
            case Cave:
                return true;
            case Water:
                return false;
            case Beach:
                // todo: expand? I think a 2x2 radius is much better in a world where you can't move water sources.
                return (type == Rock.Type.DIRT || type == Rock.Type.GRASS || type == Rock.Type.SAND || type == Rock.Type.DRY_GRASS) && // todo: dry grass?
                    (BlocksTFC.isWater(world.func_180495_p(pos.func_177982_a(1, 0, 0))) ||
                        BlocksTFC.isWater(world.func_180495_p(pos.func_177982_a(-1, 0, 0))) ||
                        BlocksTFC.isWater(world.func_180495_p(pos.func_177982_a(0, 0, 1))) ||
                        BlocksTFC.isWater(world.func_180495_p(pos.func_177982_a(0, 0, -1))));
            case Nether:
                return false;
        }

        return false;
    }
}
