/*
 * Work under Copyright. Licensed under the EUPL.
 * See the project README.md and LICENSE.txt for more information.
 */

package net.dries007.tfc.objects.container;

import java.util.stream.IntStream;
import javax.annotation.Nonnull;

import net.minecraft.entity.player.EntityPlayer;
import net.minecraft.entity.player.InventoryPlayer;
import net.minecraft.inventory.Container;
import net.minecraft.inventory.IContainerListener;
import net.minecraft.inventory.Slot;
import net.minecraft.item.ItemStack;
import net.minecraftforge.fml.relauncher.Side;
import net.minecraftforge.fml.relauncher.SideOnly;

import net.dries007.tfc.objects.te.TEInventory;
import net.dries007.tfc.util.ITileFields;

/**
 * This is the mother of all Container-with-a-Tile-Entity implementations
 *
 * @param <T> The Tile Entity class
 */
public abstract class ContainerTE<T extends TEInventory> extends Container
{
    protected final T tile;
    protected final EntityPlayer player;

    private final boolean shouldSyncCaps;
    private final boolean shouldSyncFields;
    private final int yOffset; // The number of pixels higher than normal (If the gui is larger than normal, see Anvil)

    private int[] cachedFields;

    ContainerTE(InventoryPlayer playerInv, T tile)
    {
        this(playerInv, tile, false, 0);
    }

    ContainerTE(InventoryPlayer playerInv, T tile, boolean shouldSyncCaps)
    {
        this(playerInv, tile, shouldSyncCaps, 0);
    }

    ContainerTE(InventoryPlayer playerInv, T tile, boolean shouldSyncCaps, int yOffset)
    {
        this.tile = tile;
        this.player = playerInv.field_70458_d;
        this.shouldSyncCaps = shouldSyncCaps;
        this.shouldSyncFields = tile instanceof ITileFields;
        this.yOffset = yOffset;

        addContainerSlots();
        addPlayerInventorySlots(playerInv);
    }

    @Override
    @Nonnull
    public ItemStack func_82846_b(EntityPlayer player, int index)
    {
        // Instead of overriding this, consider overriding transferIntoContainer if that is the only necessary change

        // Slot that was clicked
        Slot slot = field_75151_b.get(index);
        if (slot == null || !slot.func_75216_d())
            return ItemStack.field_190927_a;

        ItemStack stack = slot.func_75211_c();
        ItemStack stackCopy = stack.func_77946_l();

        // Transfer out of the container
        int containerSlots = field_75151_b.size() - player.field_71071_by.field_70462_a.size();
        if (index < containerSlots)
        {
            if (!this.func_75135_a(stack, containerSlots, field_75151_b.size(), true))
            {
                return ItemStack.field_190927_a;
            }
            tile.setAndUpdateSlots(index);
        }
        // Transfer into the container
        else
        {
            for (int i : getSlotShiftOrder(containerSlots))
            {
                if (field_75151_b.get(i).func_75214_a(stack))
                {
                    this.func_75135_a(stack, i, i + 1, false);
                }
            }
        }

        if (stack.func_190916_E() == 0)
        {
            slot.func_75215_d(ItemStack.field_190927_a);
        }
        else
        {
            slot.func_75218_e();
        }
        if (stack.func_190916_E() == stackCopy.func_190916_E())
        {
            return ItemStack.field_190927_a;
        }
        slot.func_190901_a(player, stack);
        return stackCopy;
    }

    @Override
    public boolean func_75145_c(@Nonnull EntityPlayer player)
    {
        return true;
    }

    protected abstract void addContainerSlots();

    @Override
    public void func_75142_b()
    {
        if (shouldSyncFields)
        {
            detectAndSendFieldChanges();
        }
        if (shouldSyncCaps)
        {
            detectAndSendAllChanges();
        }
        else
        {
            super.func_75142_b();
        }
    }

    protected void detectAndSendAllChanges()
    {
        for (int i = 0; i < field_75151_b.size(); ++i)
        {
            ItemStack stack = field_75151_b.get(i).func_75211_c();
            ItemStack newStack = field_75153_a.get(i);

            if (!ItemStack.func_77989_b(newStack, stack))
            {
                newStack = stack.func_190926_b() ? ItemStack.field_190927_a : stack.func_77946_l();
                field_75153_a.set(i, newStack);

                for (IContainerListener listener : field_75149_d)
                {
                    listener.func_71111_a(this, i, newStack);
                }
            }
        }
    }

    @Override
    @SideOnly(Side.CLIENT)
    public void func_75137_b(int id, int data)
    {
        if (shouldSyncFields)
        {
            ((ITileFields) tile).setField(id, data);
        }
    }

    protected void detectAndSendFieldChanges()
    {
        ITileFields tileFields = (ITileFields) tile;
        boolean allFieldsHaveChanged = false;
        boolean[] fieldHasChanged = new boolean[tileFields.getFieldCount()];

        if (cachedFields == null)
        {
            cachedFields = new int[tileFields.getFieldCount()];
            allFieldsHaveChanged = true;
        }

        for (int i = 0; i < cachedFields.length; ++i)
        {
            if (allFieldsHaveChanged || cachedFields[i] != tileFields.getField(i))
            {
                cachedFields[i] = tileFields.getField(i);
                fieldHasChanged[i] = true;
            }
        }

        // go through the list of listeners (players using this container) and update them if necessary
        for (IContainerListener listener : this.field_75149_d)
        {
            for (int fieldID = 0; fieldID < tileFields.getFieldCount(); ++fieldID)
            {
                if (fieldHasChanged[fieldID])
                {
                    // Note that although sendWindowProperty takes 2 ints on a server these are truncated to shorts
                    listener.func_71112_a(this, fieldID, cachedFields[fieldID]);
                }
            }
        }
    }

    private void addPlayerInventorySlots(InventoryPlayer playerInv)
    {
        // Add Player Inventory Slots
        for (int i = 0; i < 3; i++)
        {
            for (int j = 0; j < 9; j++)
            {
                func_75146_a(new Slot(playerInv, j + i * 9 + 9, 8 + j * 18, 84 + i * 18 + yOffset));
            }
        }

        for (int k = 0; k < 9; k++)
        {
            func_75146_a(new Slot(playerInv, k, 8 + k * 18, 142 + yOffset));
        }
    }

    /**
     * Helper method for overriding just container insertion priority
     */
    protected int[] getSlotShiftOrder(int containerSlots)
    {
        return IntStream.range(0, containerSlots).toArray();
    }

}
